package net.sf.sido.pojo.tool;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import net.sf.sido.*;
import net.sf.sido.array.Array;
import net.sf.sido.factory.UniversalDataFactoryReader;
import net.sf.sido.pojo.DOFactory;
import net.sf.sido.pojo.DOFactoryLogger;
import net.sf.sido.pojo.DOSchema;
import net.sf.sido.pojo.tool.model.JClass;
import net.sf.sido.pojo.tool.model.JConstructor;
import net.sf.sido.support.DataFactories;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.ClassUtils;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

public class Tool {

    private final ClassLoader classLoader;
    
    private Set<File> sourceFiles;
    private File outputDirectory;
    private File servicesDirectory;
    private String outputPackage;

    public Tool() {
        this(Tool.class.getClassLoader());
    }

    public Tool(ClassLoader classLoader) {
        this.classLoader = classLoader;
    }

    public Set<File> getSourceFiles() {
        return sourceFiles;
    }

    public void setSourceFiles(Set<File> sourceFiles) {
        this.sourceFiles = sourceFiles;
    }

    public File getOutputDirectory() {
        return outputDirectory;
    }

    public void setOutputDirectory(File outputDirectory) {
        this.outputDirectory = outputDirectory;
    }

    public String getOutputPackage() {
        return outputPackage;
    }

    public void setOutputPackage(String outputPackage) {
        this.outputPackage = outputPackage;
    }

    public File getServicesDirectory() {
        return servicesDirectory;
    }

    public void setServicesDirectory(File servicesDirectory) {
        this.servicesDirectory = servicesDirectory;
    }

    public void generate(final ToolListener listener) throws IOException {
        // Log
        listener.log("Current class loader is: %s", Thread.currentThread().getContextClassLoader());

        // Creates a factory for all the SiDOL files
        listener.log("Creating the DataFactory...");
        DataFactory dataFactory = DataFactories.typedFactory();

        // Converts files to URLs
        Collection<URL> urls = Collections2.transform(sourceFiles,
                new Function<File, URL>() {
                    @Override
                    public URL apply(File file) {
                        try {
                            return file.toURI().toURL();
                        } catch (MalformedURLException e) {
                            throw new IllegalStateException("Cannot get file URL for " + file, e);
                        }
                    }
                });

        // Logger wrapper
        DOFactoryLogger factoryLogger = new DOFactoryLogger() {
            @Override
            public void log(String pattern, Object... parameters) {
                listener.log(pattern, parameters);
            }
        };

        // Loads and discovers other schemas
        listener.log("Discovering other schemas...");
        DOFactory doFactory = new DOFactory(dataFactory, true, classLoader, factoryLogger);
        listener.log("DOFactory class loader is: %s", doFactory.getClass().getClassLoader());

        // Logs the DataSchemas
        listener.log("Imported schemas are:");
        Collection<DataSchema> schemas = doFactory.getDataFactory().getSchemas();
        for (DataSchema schema : schemas) {
            listener.log(" * %s", schema.getName());
        }

        // Logs the DO schemas
        listener.log("Imported DO schemas are:");
        Collection<DOSchema> doSchemas = doFactory.getSchemas();
        for (DOSchema doSchema : doSchemas) {
            listener.log(" * %s -> %s", doSchema.getDataSchema().getName(), doSchema.getClass().getName());
        }

        // Loads specified files
        UniversalDataFactoryReader reader = new UniversalDataFactoryReader();
        Collection<DataSchema> dataSchemas = reader.read(urls, dataFactory);

        // All classes to generate
        Set<JClass> jClasses = new HashSet<JClass>();

        // Generates all schemas and types
        Collection<JClass> schemaClasses = generate(listener, jClasses, dataFactory);

        // Writes all classes
        writeClasses(listener, jClasses);

        // Registers the schemas
        File serviceFile = new File(servicesDirectory, "META-INF/sido/sido.pojo");
        serviceFile.getParentFile().mkdirs();
        PrintWriter writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(serviceFile), "UTF-8"));
        try {
            for (JClass schemaClass : schemaClasses) {
                writer.format("%s = %s%n", schemaClass.getInfo(), schemaClass.getQualifiedName());
            }
        } finally {
            writer.close();
        }

        // Copies all schemas
        Iterator<DataSchema> dataSchemaIterator = dataSchemas.iterator();
        for (URL url : urls) {
            DataSchema dataSchema = dataSchemaIterator.next();
            // Creates the SiDOL file
            File sidolFile = new File (servicesDirectory, String.format("META-INF/sido/%s.sidol", dataSchema.getName()));
            FileUtils.copyURLToFile(url, sidolFile);
        }
    }

    private void writeClasses(ToolListener listener, Set<JClass> jClasses) throws IOException {
        for (JClass jClass : jClasses) {
            writeClass(listener, jClass);
        }
    }

    private void writeClass(ToolListener listener, JClass jClass) throws IOException {
        String className = jClass.getName();
        String fileName = className + ".java";
        String packagePath = StringUtils.replace(outputPackage, ".", "/");
        File packageFile = new File(outputDirectory, packagePath);
        if (!packageFile.exists()) {
            packageFile.mkdirs();
        }
        File file = new File(packageFile, fileName);
        PrintWriter writer = new PrintWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF-8"));
        try {
            jClass.write(writer);
        } finally {
            writer.close();
        }
    }

    private Collection<JClass> generate(ToolListener listener, Set<JClass> jClasses, DataFactory dataFactory) {
        Collection<JClass> schemaClasses = new ArrayList<JClass>();
        // Gets all schemas
        Collection<DataSchema> schemas = dataFactory.getSchemas();
        for (DataSchema schema : schemas) {
            JClass schemaClass = generateSchema(listener, jClasses, schema);
            schemaClasses.add(schemaClass);
        }
        // OK
        return schemaClasses;
    }

    private JClass generateSchema(ToolListener listener, Set<JClass> jClasses, DataSchema schema) {
        String name = schema.getName();
        String className = name + "Schema";
        log(listener, "Generating schema %s", name);
        // Class
        JClass j = new JClass(outputPackage, className);
        // TODO Header
        // Imports
        j.addImport(DataObject.class);
        j.addImport(DOFactory.class);
        j.addImport(DOSchema.class);
        // Parent
        j.setParent("DOSchema");
        // Doc
        j.addDoc("Schema for %s", name);
        j.setInfo(name);
        // Constructor
        JConstructor jConstructor = j.addConstructor()
                .addParam("DOFactory", "doFactory")
                .addContent("super(doFactory, \"%s\");", name);
        // All types
        for (DataType type : schema.getTypes()) {
            if (!type.isAbstractType()) {
                String typeName = type.getName();
                // Registration of the type in the constructor
                jConstructor.addContent("registerClass(\"%1$s::%2$s\", %1$s%2$s.class);", name, typeName);
                // Simple creation
                j.addMethod("create" + typeName, name + typeName)
                    .addContent("return new %1$s%2$s(this, newInstance(\"%1$s::%2$s\"));", name, typeName);
                // Creation from JSON
                j.addMethod("create" + typeName, name + typeName)
                    .addParam("String", "json")
                    .addContent("DataObject o = read(\"%s::%s\", json);", name, typeName)
                    .addContent("return create(o);");
            }
        }
        // OK
        jClasses.add(j);
        // Generates the types
        for (DataType type : schema.getTypes()) {
            generateType(listener, jClasses, type);
        }
        // OK
        return j;
    }

    private void generateType(ToolListener listener, Set<JClass> jClasses, DataType type) {
        String schema = type.getSchema().getName();
        String name = type.getName();
        String className = schema + name;
        log(listener, "Generating type %s", className);
        // Class
        JClass j = new JClass(outputPackage, className);
        // TODO Header
        // Imports
        j.addImport("net.sf.sido.DataObject");
        // Type class
        String parentTypeName;
        // Inheritance
        DataType parentType = type.getParentType();
        String superSchema = "schema";
        if (parentType != null) {
            parentTypeName = String.format("%s%s", parentType.getSchema().getName(), parentType.getName());
            // Gets the schema of the parent
            DataSchema parentSchema = parentType.getSchema();
            // If another schema
            if (!parentSchema.getName().equals(schema)) {
                superSchema = String.format("schema.getSchema(%sSchema.class)", parentSchema.getName());
                // FIXME Missing import for the parent schema
            }
        } else {
            // Default parent
            parentTypeName = "DOObject";
            // Imports the parent
            j.addImport("net.sf.sido.pojo.DOObject");
        }
        // Abstract class
        if (type.isAbstractType()) {
            j.setAbstractClass(true);
        }
        // Parent type
        j.setParent(parentTypeName);
        // Class doc
        j.addDoc("POJO class for %s::%s", schema, name);
        // Constructor
        j.addConstructor()
            .addParam(schema + "Schema", "schema")
            .addParam("DataObject", "o")
            .addContent("super(%s,o);", superSchema);
        // Methods
        j.addMethod(String.format("get%sSchema", schema), schema + "Schema")
            .addContent("return getSchema(%sSchema.class);", schema);
        // All properties
        Collection<Property<?>> properties = type.getProperties();
        for (Property<?> property : properties) {
            generateProperty(listener, j, property);
        }
        // End of the type class
        jClasses.add(j);
    }

    protected <T> void generateProperty(ToolListener listener, JClass j, Property<T> property) {
        String propertyName = property.getName();
        if (property.isSimple()) {
            if (property.isArray()) {
                // Array + Simple
                Class<?> pType = property.getMetaInfo().getType(PropertyMetaInfos.ELEMENT_TYPE);
                String pTypeName = pType.getSimpleName();
                String arrayTypeName = String.format("Array<%s>", pTypeName);
                // Imports the type
                j.addImport(pType);
                j.addImport(Array.class);
                // Getter
                j.addMethod(getGetMethodName(propertyName), arrayTypeName)
                    .addContent("return o.getArray(\"%s\");", propertyName);
                // Setter
                j.addMethod(getSetMethodName(propertyName))
                    .addParam(arrayTypeName, "values")
                    .addContent("o.setArray(\"%s\", values);", propertyName);
                // Add
                j.addMethod(getAddMethodName(propertyName))
                    .addParam(String.format("%s...", pTypeName), "values")
                    .addContent("o.add(\"%s\", values);", propertyName);
            } else {
                // Simple
                Class<? super T> pType = property.getType();
                // Converts to primitive type is not nullable
                String pTypeName;
                if (!property.isNullable()) {
                    // Gets the primitive type
                    Class<?> primitiveType = getPrimitiveType(pType);
                    if (primitiveType != null) {
                        pTypeName = primitiveType.getSimpleName();
                    } else {
                        pTypeName = pType.getSimpleName();
                        // Imports this type (only if not primitive)
                        j.addImport(pType);
                    }
                } else {
                    pTypeName = pType.getSimpleName();
                    // Imports this type (only if not primitive)
                    j.addImport(pType);
                }
                // Getter
                j.addMethod(getGetMethodName(propertyName), pTypeName)
                    .addContent("return o.get%s(\"%s\");", pType.getSimpleName(), propertyName);
                // Setter
                j.addMethod(getSetMethodName(propertyName))
                    .addParam(pTypeName, "value")
                    .addContent("o.set%s(\"%s\", value);", pType.getSimpleName(), propertyName);
            }
        } else {
            if (property.isArray()) {
                // Array + DO
                DataType doType = property.getMetaInfo().getObject(PropertyMetaInfos.ELEMENT_TYPE, null);
                // FIXME Array of anonymous
                if (doType == null) {

                }
                // Array of DO
                else {
                    j.addImport("net.sf.sido.array.Array");
                    j.addImport("net.sf.sido.pojo.DOUtils");
                    j.addImport("net.sf.sido.pojo.DOObject");
                    JClass doClass = getClass(doType);
                    j.addImport(doClass);
                    // Getter
                    j.addMethod(getGetMethodName(propertyName), String.format("Array<%s>", doClass.getName()))
                        .addContent("Array<DataObject> array = o.getDOArray(\"%s\");", propertyName)
                        .addContent("return array.transform(getSchema().<%s>createFn());", doClass.getName());
                    // Setter
                    j.addMethod(getSetMethodName(propertyName))
                        .addParam(String.format("Array<%s>", doClass.getName()), "values")
                        .addContent("o.setArray(\"%s\" ,values.transform(DOObject.<%s>getDataObjectFn()));", propertyName, doClass.getName());
                    // Add
                    j.addMethod(getAddMethodName(propertyName))
                        .addParam(String.format("%s...", doClass.getName()), "values")
                        .addContent("o.add(\"%s\", DOUtils.transform(values));", propertyName);
                }
            } else {
                // DO
                DataType doType = property.getMetaInfo().getObject(PropertyMetaInfos.DO_TYPE, null);
                // Anonymous
                if (doType == null) {
                    j.addImport("net.sf.sido.pojo.DOObject");
                    j.addMethod(getGetMethodName(propertyName), "DOObject")
                        .addContent("return getSchema().create(o.getDataObject(\"%s\"));", propertyName);
                    j.addMethod(getSetMethodName(propertyName))
                        .addParam("DOObject", "value")
                        .addContent("o.setDataObject(\"%s\", value.getDataObject());", propertyName);
                }
                // DO type
                else {
                    j.addMethod(getGetMethodName(propertyName), getClass(doType))
                        .addContent("return getSchema().create(o.getDataObject(\"%s\"));", propertyName);
                    j.addMethod(getSetMethodName(propertyName))
                        .addParam(getClass(doType), "value")
                        .addContent("o.setDataObject(\"%s\", value.getDataObject());", propertyName);
                }
            }
        }
    }

    private Class<?> getPrimitiveType(Class<?> type) {
        return ClassUtils.wrapperToPrimitive(type);
    }

    private JClass getClass(DataType type) {
        // FIXME Other schemas
        return new JClass(outputPackage, type.getSchema().getName() + type.getName());
    }

    private String getSetMethodName(String name) {
        return getAccessorName("set", name);
    }

    private String getAccessorName(String prefix, String name) {
        return String.format("%s%s", prefix, StringUtils.capitalize(name));
    }

    private String getGetMethodName(String name) {
        return getAccessorName("get", name);
    }

    private String getAddMethodName(String name) {
        return getAccessorName("add", name);
    }

    private void log(ToolListener listener, String pattern, Object... params) {
        listener.log(pattern, params);
    }
}
